import argparse
import json
import logging

from quollio_core.helper.env_default import env_default
from quollio_core.helper.log import set_log_level
from quollio_core.profilers.bigquery import bigquery_table_lineage
from quollio_core.repository import qdc
from quollio_core.repository.bigquery import get_credentials, get_org_id

logger = logging.getLogger(__name__)


def load_lineage(
    qdc_client: qdc.QDCExternalAPIClient, project_id: str, regions: list, tenant_id: str, credentials: dict, org_id: str
):
    bigquery_table_lineage(
        qdc_client=qdc_client,
        tenant_id=tenant_id,
        project_id=project_id,
        regions=regions,
        credentials=credentials,
        org_id=org_id,
    )


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        prog="Quollio Intelligence Agent for Google BigQuery",
        description="Collect lineage and stats from Google BigQuery and load to Quollio Data Catalog",
        epilog="Copyright (c) 2024 Quollio Technologies, Inc.",
    )
    parser.add_argument(
        "commands",
        choices=["load_lineage"],
        type=str,
        nargs="+",
        help="""
        The command to execute.
        'load_lineage': Load lineage data from Google Data Catalog to Quollio,
        """,
    )
    parser.add_argument(
        "--credentials",
        type=str,
        action=env_default("GOOGLE_APPLICATION_CREDENTIALS"),
        help="Crendentials for Google Cloud Platform",
    )
    parser.add_argument(
        "--tenant_id",
        type=str,
        action=env_default("TENANT_ID"),
        required=False,
        help="The tenant id (company id) where the lineage and stats are loaded",
    )
    parser.add_argument(
        "--api_url",
        type=str,
        action=env_default("QDC_API_URL"),
        required=False,
        help="The base URL of Quollio External API",
    )
    parser.add_argument(
        "--client_id",
        type=str,
        action=env_default("QDC_CLIENT_ID"),
        required=False,
        help="The client id that is created on Quollio console to let clients access Quollio External API",
    )
    parser.add_argument(
        "--client_secret",
        type=str,
        action=env_default("QDC_CLIENT_SECRET"),
        required=False,
        help="The client secret that is created on Quollio console to let clients access Quollio External API",
    )
    parser.add_argument(
        "--project_id",
        type=str,
        action=env_default("GCP_PROJECT_ID"),
        required=False,
        help="GCP Project ID",
    )
    parser.add_argument(
        "--regions",
        type=str,
        action=env_default("GCP_REGIONS"),
        required=False,
        help="GCP regions where the data is located. Multiple regions can be provided separated by space.",
        nargs="+",
    )
    parser.add_argument(
        "--log_level",
        type=str,
        choices=["debug", "info", "warn", "error", "none"],
        action=env_default("LOG_LEVEL"),
        required=False,
        help="The log level for dbt commands. Default value is info",
    )

    args = parser.parse_args()
    set_log_level(level=args.log_level)

    if len(args.commands) == 0:
        raise ValueError("No command is provided")

    if "load_lineage" in args.commands:
        qdc_client = qdc.QDCExternalAPIClient(
            base_url=args.api_url, client_id=args.client_id, client_secret=args.client_secret
        )

        credentials_json = json.loads(args.credentials)
        credentials = get_credentials(credentials_json=credentials_json)
        org_id = get_org_id(credentials_json=credentials_json)

        load_lineage(
            qdc_client=qdc_client,
            project_id=args.project_id,
            regions=args.regions,
            tenant_id=args.tenant_id,
            credentials=credentials,
            org_id=org_id,
        )



import logging

from google.cloud.bigquery import Client
from google.cloud.datacatalog_lineage_v1 import EntityReference, LineageClient, SearchLinksRequest
from google.oauth2.service_account import Credentials
from googleapiclient.discovery import build

logger = logging.getLogger(__name__)


class BigQueryClient:
    def __init__(self, credentials: Credentials) -> None:
        self.client = self.__initialze(credentials=credentials)

    def __initialze(self, credentials: Credentials) -> Client:
        client = Client(credentials=credentials)
        return client

    def list_datasets(self, project_id) -> list:
        datasets = list(self.client.list_datasets(project_id))
        logger.debug("Found %s datasets in project %s", len(datasets), project_id)
        return datasets

    def list_tables(self, dataset_id) -> list:
        tables = list(self.client.list_tables(dataset_id))
        logger.debug("Found %s tables in dataset %s", len(tables), dataset_id)
        return list(self.client.list_tables(dataset_id))


class GCPLineageClient:
    def __init__(self, credentials: Credentials) -> None:
        self.client = self.__initialze(credentials=credentials)

    def __initialze(self, credentials: Credentials) -> LineageClient:
        client = LineageClient(credentials=credentials)
        return client

    def get_links(self, request: SearchLinksRequest) -> list:
        response = self.client.search_links(request)
        return response.links


def get_entitiy_reference() -> EntityReference:
    return EntityReference()


def get_search_request(downstream_table: EntityReference, project_id: str, region: str) -> SearchLinksRequest:
    return SearchLinksRequest(target=downstream_table, parent=f"projects/{project_id}/locations/{region.lower()}")


def get_credentials(credentials_json: dict) -> Credentials:
    return Credentials.from_service_account_info(credentials_json)


def get_org_id(credentials_json: dict) -> str:
    credentials = get_credentials(credentials_json)
    crm_service = build("cloudresourcemanager", "v1", credentials=credentials)
    project_id = credentials_json["project_id"]
    project = crm_service.projects().get(projectId=project_id).execute()
    org_id = project["parent"]["id"]
    return org_id



import logging
from typing import Any, Dict, List

from quollio_core.profilers.lineage import gen_table_lineage_payload, parse_bigquery_table_lineage
from quollio_core.repository import qdc
from quollio_core.repository.bigquery import BigQueryClient, GCPLineageClient, get_entitiy_reference, get_search_request

logger = logging.getLogger(__name__)


def bigquery_table_lineage(
    qdc_client: qdc.QDCExternalAPIClient,
    tenant_id: str,
    project_id: str,
    regions: list,
    org_id: str,
    credentials: Any,
):
    lineage_client = GCPLineageClient(credentials)
    bq_client = BigQueryClient(credentials)

    datasets = bq_client.list_datasets(project_id)
    all_tables = generate_table_list(datasets, bq_client)
    lineage_links = generate_lineage_links(all_tables, lineage_client, project_id, regions)
    lineage_links = parse_bigquery_table_lineage(lineage_links)

    update_table_lineage_inputs = gen_table_lineage_payload(tenant_id=tenant_id, endpoint=org_id, tables=lineage_links)

    req_count = 0
    for update_table_lineage_input in update_table_lineage_inputs:
        logger.info(
            "Generating table lineage. downstream: %s -> %s-> %s",
            update_table_lineage_input.downstream_database_name,
            update_table_lineage_input.downstream_schema_name,
            update_table_lineage_input.downstream_table_name,
        )
        status_code = qdc_client.update_lineage_by_id(
            global_id=update_table_lineage_input.downstream_global_id,
            payload=update_table_lineage_input.upstreams.as_dict(),
        )
        if status_code == 200:
            req_count += 1
    logger.info("Generating table lineage is finished. %s lineages are ingested.", req_count)


def generate_table_list(datasets: List[str], bq_client: BigQueryClient) -> List[str]:
    all_tables = []
    for dataset in datasets:
        all_tables.extend(
            [
                table
                for table in bq_client.list_tables(dataset.dataset_id)
                if table.table_type in ["TABLE", "VIEW", "MATERIALIZED_VIEW"]
            ]
        )

    all_table_names = []
    for table in all_tables:
        all_table_names.append(f"{table.project}.{table.dataset_id}.{table.table_id}")

    return all_table_names


def generate_lineage_links(
    all_tables: List[str], lineage_client: GCPLineageClient, project_id: str, regions: List[str]
) -> Dict[str, List[str]]:
    lineage_links = {}
    for table in all_tables:
        downstream = get_entitiy_reference()
        downstream.fully_qualified_name = f"bigquery:{table}"

        for region in regions:
            request = get_search_request(downstream_table=downstream, project_id=project_id, region=region)
            response = lineage_client.get_links(request=request)
            for lineage in response:
                target_table = str(lineage.target.fully_qualified_name).replace("bigquery:", "")
                if target_table not in lineage_links:
                    lineage_links[target_table] = []
                lineage_links[target_table].append(str(lineage.source.fully_qualified_name).replace("bigquery:", ""))

    return lineage_links
